@@ -2,7 +2,7 @@ package Pod::WikiDoc;
 use strict;
 use warnings;
 use vars qw($VERSION );
-$VERSION     = "0.17";
+$VERSION     = '0.18';
 
 use 5.006;
 use Carp;
@@ -253,7 +253,7 @@ sub filter {
 ### translation process or who want to convert wikidoc to Pod for other
 ### creative purposes using the Pod::WikiDoc engine.
 
-sub format {
+sub format { ## no critic
     my ($self, $wikitext) = @_;
     
     croak "Error: Argument to format() must be a scalar"
@@ -280,7 +280,80 @@ sub format {
 sub _comment_block_regex {
     my ( $self ) = @_;
     my $length = $self->{comment_prefix_length};
-    return "\\A#{$length} (.*)?\\z";
+    return qr/\A#{$length}(?:\s(.*))?\z/ms;
+}
+
+#--------------------------------------------------------------------------#
+# _input_iterator
+#
+# return an iterator that streams a filehandle. Action arguments:
+#     'peek' -- lookahead at the next line without consuming it
+#     'next' and 'drop' -- synonyms to consume and return the next line
+#--------------------------------------------------------------------------#
+
+sub _input_iterator {
+    my ($self, $fh) = @_;
+    my @head;
+    return sub {
+        my ($action) = @_;
+        if ($action eq 'peek') {
+            push @head, scalar <$fh> unless @head;
+            return $head[0];
+        }
+        elsif ( $action eq 'drop' || $action eq 'next' ) {
+            return shift @head if @head;
+            return scalar <$fh>;
+        }
+        else {
+            croak "Unrecognized iterator action '$action'\n";
+        }
+    }
+}
+
+#--------------------------------------------------------------------------#
+# _exhaust_iterator
+#
+# needed to help abort processing
+#--------------------------------------------------------------------------#
+
+sub _exhaust_iterator {
+    my ($self, $iter) = @_;
+    1 while $iter->();
+    return;
+}
+
+#--------------------------------------------------------------------------#
+# _output_iterator
+#
+# returns an output "iterator" that streams to a filehandle.  Inputs
+# are array refs of the form [ $FORMAT, @LINES ].  Format 'pod' is 
+# printed to the filehandle immediately.  Format 'wikidoc' is accumulated
+# until the next 'pod' then converted to wikidoc and printed to the file
+# handle
+#--------------------------------------------------------------------------#
+
+sub _output_iterator {
+    my ($self, $fh) = @_;
+    my @wikidoc;
+    return sub {
+        my ($chunk) = @_;
+        if ($chunk eq 'flush') {
+            print {$fh} $self->format( join(q{}, splice(@wikidoc,0) ) )
+                if @wikidoc;
+            return;
+        }
+        return unless ref($chunk) eq 'ARRAY';
+        my ($format, @lines) = @$chunk;
+        if ( $format eq 'wikidoc' ) {
+            push @wikidoc, @lines;
+        }
+        elsif ( $format eq 'pod' ) {
+            print {$fh} $self->format( join(q{}, splice(@wikidoc,0) ) )
+                if @wikidoc;
+            print {$fh} @lines;
+        }
+        return;
+    }
 }
 
 #--------------------------------------------------------------------------#
@@ -290,119 +363,191 @@ sub _comment_block_regex {
 # markup to Pod in the process
 #--------------------------------------------------------------------------#
 
+my $BLANK_LINE = qr{\A \s* \z}xms;
+my $NON_BLANK_LINE = qr{\A \s* \S }xms;
+my $FORMAT_LABEL = qr{:? [-a-zA-Z0-9_]+}xms;
+my $POD_CMD = qr{\A =[a-zA-Z]+}xms;
+my $BEGIN = qr{\A =begin \s+ ($FORMAT_LABEL)  \s* \z}xms;
+my $END   = qr{\A =end   \s+ ($FORMAT_LABEL)  \s* \z}xms;
+my $FOR   = qr{\A =for   \s+ ($FORMAT_LABEL)  [ \t]* (.*) \z}xms;
+my $POD   = qr{\A =pod                          \s* \z}xms;
+my $CUT   = qr{\A =cut                          \s* \z}xms;
+
 sub _filter_podfile {
     my ($self, $input_fh, $output_fh) = @_;
 
-    # initialize flags and buffers
-    my $in_pod      = 0; # not in a Pod section at start
-    my $in_begin    = 0; # not in a begin section 
-    my $in_wikidoc  = 0; # not in a wikidoc section
-    my $in_comment_block = 0; # not in a wikidoc comment section
-    my @wikidoc;
-    
-    # init regex
-    my $comment_regex = _comment_block_regex($self);
-    my $MATCHES_COMMENT_BLOCK = qr{$comment_regex}ms;
-    
     # open output with tag and Pod marker
     print $output_fh
           "# Generated by Pod::WikiDoc version $VERSION\n\n";
     print $output_fh "=pod\n\n";
     
-    # process line-by-line
-    my $line;
-    LINE:
-    while ( defined( $line = <$input_fh> ) ) {
-        if ( $in_pod && ( ! $in_wikidoc) ) {
-            if ( $line =~ m{ \A =cut }xms ) {
-                $in_pod = 0;
-                next LINE;
-            }
-            elsif ( $line =~ m{ \A =(begin|for) \s+ wikidoc \s* (.*)}xms ) {
-                my ($command, $para) = ($1, $2);
-                
-                $in_wikidoc = 1;
-                $in_begin = $command eq 'begin' ? 1 : 0;
-                
-                # if =for wikidoc, then store $para
-                if ( ! $in_begin && defined $para && length $para ) {
-                    push @wikidoc, $para;
-                }
-                next LINE;
-            }
-            else {
-                print $output_fh $line;
-            }
-        }
-        elsif ( $in_pod && $in_wikidoc ) {
-            # see if we're done -- =begin/=end or =for/blankline
-            if (    (   $in_begin && $line =~ m{\A =end \s+ wikidoc}xms )
-                 || ( ! $in_begin && $line =~ m{\A \s*  \z         }xms ) ) {
-                
-                print $output_fh _translate_wikidoc( $self, \@wikidoc );
-                @wikidoc = ();
-                $in_wikidoc = $in_begin = 0;
-                next LINE;
-            }
-            # not done, so store up the wikidoc
-            push @wikidoc, $line;
-            next LINE;
-        }
-        elsif ( ( ! $in_pod ) && $in_comment_block ) {
-            # capture a comment_block line
-            if ( $line =~ $MATCHES_COMMENT_BLOCK ) {
-                my $comment_block_text = defined $1 ? $1 : q{};
-                push @wikidoc, $comment_block_text; 
-                next LINE;
-            }
-            else { 
-                print $output_fh _translate_wikidoc( $self, \@wikidoc );
-                @wikidoc = ();
-                $in_comment_block = 0;
-                redo LINE;
-            }
-        }
-        else {
-            # if it's a Pod command, flag and handle it
-            if ( $line =~ m{ \A = ([a-zA-Z]\S*) }xms ) {
-                my $command = $1;
-                
-                # =cut can't start pod
-                if ( $command eq "cut" ) {
-                    next LINE;
-                }
+    # setup iterators
+    my $in_iter = $self->_input_iterator( $input_fh );
+    my $out_iter = $self->_output_iterator( $output_fh );
 
-                # flag that we've found Pod
-                $in_pod = 1;
-
-                # eat a =pod marker, otherwise, process it as Pod
-                if ( $command eq "pod" ) {
-                    next LINE;
-                }
-                else {
-                    redo LINE;
-                }
-            }
-            # if it's a comment block and we're using them, flag and restart
-            elsif ( $self->{comment_blocks} && $line =~ $MATCHES_COMMENT_BLOCK ) {
-                $in_comment_block = 1;
-                redo LINE;
-            }
-            # otherwise, move on
-            else {
-                next LINE;
-            }
-        }
-    } # while
-    
-    # print any unfinished wikidoc capture
-    if ( @wikidoc ) {
-        print $output_fh _translate_wikidoc( $self, \@wikidoc );
-    }
+    # starting filter mode is code
+    $self->_filter_code( $in_iter, $out_iter );
+    $out_iter->('flush');
 
     return;
 }
 
+#--------------------------------------------------------------------------#
+# _filter_code
+#
+# we need a "cutting" flag -- if we got here from a =cut, then we return to 
+# caller ( pod or format ) when we see pod. Otherwise we're just starting
+# and need to start a new pod filter when we see pod
+#
+# perlpodspec says starting Pod with =cut is an error and that we 
+# *must* halt parsing and *should* issue a warning. Here we might be
+# far down the call stack and don't want to just return where the caller
+# might continue processing.  To avoid this, we exhaust the input first.
+#--------------------------------------------------------------------------#
+
+sub _filter_code {
+    my ($self, $in_iter, $out_iter, $cutting) = @_;
+    my $CBLOCK = _comment_block_regex($self);
+    CODE: while ( defined( my $peek = $in_iter->('peek') ) ) {
+        $peek =~ $CBLOCK && do {
+            $self->_filter_cblock( $in_iter, $out_iter );
+            next CODE;
+        };
+        $peek =~ $CUT && do {
+            warn "Can't start Pod with '$peek'\n";
+            $self->_exhaust_iterator( $in_iter );
+            last CODE;
+        };
+        $peek =~ $POD_CMD && do {
+            last CODE if $cutting;
+            $self->_filter_pod( $in_iter, $out_iter );
+            next CODE;
+        };
+        do { $in_iter->('drop') };
+    }
+    return;
+}
+
+#--------------------------------------------------------------------------#
+# _filter_pod
+#
+# Pass through lines to the output iterators, but flag wikidoc lines 
+# differently so that they can be converted on output
+#
+# If we find an =end that is out of order, perlpodspec says we *must* warn
+# and *may* halt.  Instead of halting, we return to the caller in the
+# hopes that an earlier format might match this =end.
+#--------------------------------------------------------------------------#
+
+sub _filter_pod {
+    my ($self, $in_iter, $out_iter) = @_;
+    my @format = (); # no format to start
+    # process the pod block -- recursing as necessary
+    LINE: while ( defined( my $peek = $in_iter->('peek') ) ) {
+        $peek =~ $POD && do {
+            $in_iter->('drop');
+            next LINE;
+        };
+        $peek =~ $CUT && do { 
+            $in_iter->('drop');
+            $self->_filter_code( $in_iter, $out_iter, 1 );
+            next LINE;
+        };
+        $peek =~ $FOR && do {
+            $self->_filter_for( $in_iter, $out_iter );
+            next LINE;
+        };
+        $peek =~ $END && do {
+            if ( ! @format ) {
+                warn "Error: '$peek' doesn't match any '=begin $1'\n";
+                $in_iter->('drop');
+                next LINE;
+            }
+            elsif ( $format[-1] ne $1 ) {
+                warn "Error: '$peek' doesn't match '=begin $format[-1]'\n";
+                pop @format; # try an earlier format
+                redo LINE;
+            }
+            elsif ( $format[-1] eq 'wikidoc' ) {
+                pop @format;
+                $in_iter->('drop');
+                next LINE;
+            }
+            else {
+                pop @format;
+                # and let it fall through to the output iterator
+            }
+        };
+        $peek =~ $BEGIN && do {
+            if ( $1 eq 'wikidoc' ) {
+                push @format, 'wikidoc';
+                $in_iter->('drop');
+                next LINE;
+            }
+            else {
+                push @format, $1;
+                # and let it fall through to the output iterator
+            }
+        };
+        do { 
+            my $out_type = 
+                ( @format && $format[-1] eq 'wikidoc' ) ? 'wikidoc' : 'pod' ; 
+            $out_iter->( [ $out_type, $in_iter->('next') ] ) 
+        };
+    }
+    return;
+}
+
+#--------------------------------------------------------------------------#
+# _filter_for
+#--------------------------------------------------------------------------#
+
+sub _filter_for {
+    my ($self, $in_iter, $out_iter) = @_;
+    my $for_line = $in_iter->('next');
+    my ($format, $rest) = $for_line =~ $FOR;
+    $rest ||= "\n";
+
+    my @lines = ( $format eq 'wikidoc' ? $rest : $for_line );
+
+    LINE: while ( defined( my $peek = $in_iter->('peek') ) ) {
+        $peek =~ $BLANK_LINE && do {
+            last LINE;
+        };
+        do {
+            push @lines, $in_iter->('next');
+        };
+    }
+    if ($format eq 'wikidoc' ) {
+        $in_iter->('drop'); # wikidoc will append \n
+    }
+    else {
+        push @lines, $in_iter->('next');
+    }
+    my $out_type =  $format eq 'wikidoc' ? 'wikidoc' : 'pod' ; 
+    $out_iter->( [ $out_type, @lines ] ); 
+    return;
+}
+
+#--------------------------------------------------------------------------#
+# _filter_cblock
+#--------------------------------------------------------------------------#
+
+sub _filter_cblock {
+    my ($self, $in_iter, $out_iter) = @_;
+    my @lines = ($1 ? $1 : "\n"); ## no critic
+    $in_iter->('next');
+    my $CBLOCK = _comment_block_regex($self);
+    LINE: while ( defined( my $peek = $in_iter->('peek') ) ) {
+        last LINE if $peek !~ $CBLOCK;
+        push @lines, ($1 ? $1 : "\n");
+        $in_iter->('next');
+    }
+    $out_iter->( [ 'wikidoc', @lines ] ) if $self->{comment_blocks};
+    return;
+}
+
+
 #--------------------------------------------------------------------------#
 # Translation functions and tables
 #--------------------------------------------------------------------------#
@@ -469,8 +614,8 @@ my %closing_of = (
     BoldText            =>  ">",
     ItalicText          =>  ">",
     KeyWord             =>  q{},
-    LinkContent         =>  ">",
-    LinkLabel           =>  "|",
+    LinkContent         =>  q{>},
+    LinkLabel           =>  q{|},
     LinkTarget          =>  q{},
 );
 
@@ -484,10 +629,10 @@ my %content_handler_for = (
 
 # Table of character to E<> code conversion
 my %escape_code_for = (
-    ">" =>  "E<gt>",
-    "<" =>  "E<lt>",
-    "|" =>  "E<verbar>",
-    "/" =>  "E<sol>",
+    q{>} =>  "E<gt>",
+    q{<} =>  "E<lt>",
+    q{|} =>  "E<verbar>",
+    q{/} =>  "E<sol>",
 );
 
 # List of characters that need conversion
@@ -529,7 +674,7 @@ sub _keyword_expansion {
     my ($node, $keywords) = @_;
     my $key = $node->{content};
     my $value = $keywords->{$key};
-    return defined $value ? $value : "%%" . $key . "%%" ;
+    return defined $value ? $value : q{%%} . $key . q{%%} ;
 }
 
     
@@ -752,42 +897,26 @@ existing test-file that illustrates the bug or desired feature.
 
 David A. Golden (DAGOLDEN)
 
-dagolden@cpan.org
-
-http://dagolden.com/
-
 = COPYRIGHT AND LICENSE
 
-Copyright (c) 2005,2006 by David A. Golden
+Copyright (c) 2005, 2006, 2007 by David A. Golden
 
-This program is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself.
+Licensed under the Apache License, Version 2.0 (the "License");
+you may not use this file except in compliance with the License.
+You may obtain a copy of the License at 
+[http://www.apache.org/licenses/LICENSE-2.0]
 
-The full text of the license can be found in the LICENSE file included with
-this module.
+Files produced as output though the use of this software, including
+generated copies of boilerplate templates provided with this software,
+shall not be considered Derivative Works, but shall be considered the
+original work of the Licensor.
 
-= DISCLAIMER OF WARRANTY
-
-BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
-FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
-OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
-PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
-EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
-WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
-ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
-YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
-NECESSARY SERVICING, REPAIR, OR CORRECTION.
-
-IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
-WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
-REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
-LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
-OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
-THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
-RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
-FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
-SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
-SUCH DAMAGES.
+Unless required by applicable law or agreed to in writing, software
+distributed under the License is distributed on an "AS IS" BASIS,
+WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+See the License for the specific language governing permissions and
+limitations under the License.
 
 =end wikidoc
 
+=cut
